// Interpreter.cpp: implementation of the Interpreter class.
//
//////////////////////////////////////////////////////////////////////

#include "Interpreter.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Interpreter::Interpreter()
{
}

Interpreter::Interpreter(const Interpreter &right)
{
	input = right.input;
	output = right.output;
	tag_table = right.tag_table;
}

Interpreter::~Interpreter()
{
}

const Interpreter &Interpreter::operator=(const Interpreter &right)
{
	if (this != &right) {
		input = right.input;
		output = right.output;
		tag_table = right.tag_table;
	}
	return *this;
}

Document *Interpreter::getInput() const
{
	return input;
}

vector<DecoratedChar> *Interpreter::getOutput() const
{
	return output;
}

void Interpreter::setInput(Document *d)
{
	input = d;
}

void Interpreter::setOutput(vector<DecoratedChar> *vdc)
{
	output = vdc;
}

void Interpreter::setup()
{
}

void Interpreter::run()
{
	int i,j;
	bool isTagOpen;
	string tag;
	Character cha;
	PropertyRecord pr;
	DecoratedChar dc;

	const string &str = input->getDocument();
	setup();
	pr.setup();
	i = 0;
	while (i<str.length()) {
		if ( str[i] == '<') {
			i++;
			if ( str[i] == '/') {
				isTagOpen = false;
				i++;
			}
			else
				isTagOpen = true;
			j = 1;
			while (str[i+j] != '>') {
				j++;
			}
			tag = str.substr(i,j);
			if (isTagOpen == true) {
				prstack.push_back(pr);
				pr.applyTag(tag_table.findTag(tag));
			}
			else {
				pr = prstack.back();
				prstack.pop_back();
			}
			i += j;
		}
		else {
			cha.setCharacter(str[i]);
			dc.setDecoration(pr);
			dc.setCharacter(cha);
			if (dc.getCharacter().getCharacterType() == WHITE_SPACE) {
				if (dc.getDecoration().tt==true) 
					output->push_back(dc);
				else {
					cha.setCharacter(0x20);
					dc.setCharacter(cha);
					if (output->size() == 0)
						output->push_back(dc);
					else {
						if (dc!=output->back()) 
							output->push_back(dc);	
					}
				}
			}
			else 
				output->push_back(dc);
		}
		i++;
	}
}

void Interpreter::print() const
{
	vector<DecoratedChar>::iterator i;
	int j=0;

	cout<<"Interpreter Output"<<endl;
	for (i=output->begin();i!=output->end();i++) {
		cout<<j++<<" ";
		i->print();
	}
//	output->operator[](output->size()-1).print();
}

