// Optimizer.cpp: implementation of the Optimizer class.
//
//////////////////////////////////////////////////////////////////////

#include "Optimizer.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Optimizer::Optimizer()
{
}

Optimizer::Optimizer(const Optimizer &right)
{
	input = right.input;
	state_change_table = right.state_change_table;
	best_move_table = right.best_move_table;
	tag_table = right.tag_table;
}

Optimizer::~Optimizer()
{
}

const Optimizer &Optimizer::operator=(const Optimizer &right)
{
	if (this != &right) {
		input = right.input;
		state_change_table = right.state_change_table;
		best_move_table = right.best_move_table;
		tag_table = right.tag_table;
	}
	return *this;
}

vector<DecoratedChar> *Optimizer::getInput() const
{
	return input;
}

void Optimizer::setInput(vector<DecoratedChar> *vdc)
{
	input = vdc;
}

void Optimizer::setup()
{
	PropertyRecord pr;
	pr.setup();
	prstack.push_back(pr);
}

void Optimizer::run()
{
	vector<int>::size_type i,j;
	Move move;
	PropertyRecord pr;


	setup();
	makeStateChangeTable();
	for (j=0;j<state_change_table.size();j++) {
		search(j,5);
		//Make best move
		pr = prstack.back();
		for (i=0;i<last_best_move.tag_table.size();i++) {
			if (last_best_move.tag_table[i] == -1) {
				prstack.pop_back();
				pr = prstack.back();
			}
			else {
				pr.applyTag(last_best_move.tag_table[i]);
				prstack.push_back(pr);
			}
		}
		//Store best move
		best_move_table.push_back(last_best_move);
	}
	//last move:close remaining tags
	move.clear();
	for (i=1;i<prstack.size();i++)
		move.tag_table.push_back(-1);
	best_move_table.push_back(move);
}

int Optimizer::search(const int index,int depth)
{
	vector<int>::size_type i;
	vector<Move>::size_type im;
	vector<Move> move_table;
	vector<PropertyRecord> temp_prstack;
	Move best_move,move;
	PropertyRecord pr;
	int best_score,score;

	if (depth == 0)
		return 0;
	//Find possible moves
	findMoves(index,move_table);
	//Determine best move
	best_score = 99999999;
	for (im=0;im<move_table.size();im++) {
		move = move_table[im];
		score = scoreMove(move);
		temp_prstack = prstack;
		pr = prstack.back();
		for (i=0;i<move.tag_table.size();i++) {
			if (move.tag_table[i] == -1) {
				prstack.pop_back();
				pr = prstack.back();
			}
			else {
				pr.applyTag(move.tag_table[i]);
				prstack.push_back(pr);
			}
		}	
		if (state_change_table.size() != index+1)
			score+=search(index+1,depth-1);
		if (score<best_score) {
			best_score = score;
			best_move = move_table[im];
		}
		prstack = temp_prstack;
	}
	last_best_move = best_move;
	return best_score;
}

void Optimizer::findMoves(const int index,vector<Move> &move_table)
{
	int index1,popnum,i,udiff;
	Character ch1;
	DecoratedChar dc1,dc2;
	PropertyRecord currentpr,newpr;
	Move move;
	
	index1 = state_change_table[index];
	dc1 = input->operator[](index1);	
	ch1 = dc1.getCharacter();
	newpr = dc1.getDecoration();
	
	for (popnum = 0; popnum<prstack.size(); popnum++) {
		//Pop popnum off
		if (popnum>10 && move_table.size()>0)
			break;
		currentpr = prstack[prstack.size()-1-popnum];
		move.clear();
		for (i=0;i<popnum;i++) 
			move.tag_table.push_back(-1);

		//Find open tags
		udiff = newpr.u - currentpr.u;
		//Change Text
		//Need plain move if need to unset b,i,s,tt or u
		if ( (currentpr.b==true && newpr.b==false) ||
			 (currentpr.i==true && newpr.i==false) ||
		     (currentpr.s==true && newpr.s==false) ||
			 (currentpr.tt==true && newpr.tt==false) ||
			 udiff < 0 ) 
		{
			move.tag_table.push_back(3);
			currentpr.applyTag(3);
		}
		for (i=0;i<udiff;i++) {
			move.tag_table.push_back(6);
			currentpr.applyTag(6);
		}
		dc2.setDecoration(currentpr);
		dc2.setCharacter(ch1);
		currentpr = dc2.getDecoration();
		if (currentpr.b==false && newpr.b==true) { 
			move.tag_table.push_back(0);
			currentpr.applyTag(0);
		}
		if (currentpr.i==false && newpr.i==true) {
			move.tag_table.push_back(2);
			currentpr.applyTag(2);
		}
		if (currentpr.tt==false && newpr.tt==true) {
			move.tag_table.push_back(5);
			currentpr.applyTag(5);
		}
		if (currentpr.s==false && newpr.s==true) {
			move.tag_table.push_back(4);
			currentpr.applyTag(4);
		}
		if (currentpr.em==false && newpr.em==true) {
			move.tag_table.push_back(1);
			currentpr.applyTag(1);
		}
		//Change Size
		if (currentpr.size!=newpr.size && newpr.size!=ANY_SIZE ) {
			move.tag_table.push_back(newpr.size+7);
			currentpr.applyTag(newpr.size+7);
		}
		//Change color
		if (currentpr.color!=newpr.color && newpr.color!=ANY_COLOR) {
			move.tag_table.push_back(newpr.color+17);
			currentpr.applyTag(newpr.color+17);
		}
	
		if (currentpr == newpr) 
			move_table.push_back(move);	
	}
}

void Optimizer::print() const
{
	vector<int>::size_type i;

	cout<<"State Change Table "<<state_change_table.size()<<endl;
	for (i=0;i<state_change_table.size();i++) {
		cout<<i<<" "<<state_change_table[i]<<" ";
		input->operator[](state_change_table[i]).print();
	}
	cout<<"Best Move Table "<<best_move_table.size()<<endl;
	for (i=0;i<best_move_table.size();i++) {
		best_move_table[i].print();
		cout<<endl;
	}
}

void Optimizer::printResult(Document *orig_input) const
{
	int i,j,k,index1,index2;
	Move move;
	TagTable tt;
	string tag;
	vector<int> move_stack;
	string temp;
	
	for (i=0;i<best_move_table.size();i++) {
		if (i == state_change_table.size())
			index1 = input->size();
		else
			index1 = state_change_table[i];
		if (i==0)
			index2 = 0;
		else
			index2 = state_change_table[i-1];
		for (j=index2;j<index1;j++) 
			temp+=(input->operator[](j).getCharacter().getCharacter());
		move = best_move_table[i];
		for (j=0;j<move.tag_table.size();j++) {
			k = move.tag_table[j];
			if (k==-1) {
				tag = '/' + tt.getTable()[move_stack.back()].getName();
				move_stack.pop_back();
			}
			else {
				tag = tt.getTable()[k].getName();
				move_stack.push_back(k);
			}
			temp+="<";
			temp+=tag;
			temp+=">";
		}
	}
	if ( (temp.length()) < (orig_input->getDocument().length()) )
		cout << temp;
	else
		cout << orig_input->getDocument();
}

void Optimizer::makeStateChangeTable()
{
	vector<DecoratedChar>::size_type i;
	DecoratedChar dc1,dc2;
	PropertyRecord pr1;
	
	pr1.setup();
	for (i=0;i<input->size();i++) {
		dc1 = input->operator[](i);
		dc2.setDecoration(pr1);
		dc2.setCharacter(dc1.getCharacter());
		if (dc1 != dc2) {
			state_change_table.push_back(i);
			pr1 = dc1.getDecoration();
		}
	}
}

int Optimizer::scoreMove(const Move &move)
{
	int k;
	vector<int>::size_type vi;
	
	k = 0;
	for (vi=0;vi<move.tag_table.size();vi++) {
		if (move.tag_table[vi] != -1) 
			k++;
	}
	return k;
}